/*
 * loadkeys.y
 *
 * merged from version 0.81 of loadkeys.y and mktable.c - aeb@cwi.nl
 * Reason for change:
 *   The original version of mktable would use the current kernel
 *   for getting at the entries. However, this does not work if
 *   e.g. NR_FUNC in keyboard.h has changed. So, instead of getting
 *   the information from the current kernel, we must get it out of
 *   defkeymap.map, just like loadkeys. Thus, mktable is now an
 *   option of loadkeys.
 * (Other advantage: we first do the parsing, and then the key binding.
 *  No bindings are changed in case of a syntax error.)
 * Fix: contrary to the documentation in keytables.5 it was assumed
 * by default that the AltGr modifier had no effect. This assumption
 * has been removed.
 */

%token EOL NUMBER LITERAL KEYCODE EQUALS SHIFT CONTROL ALT ALTGR STRING STRLITERAL COMPOSE TO CCHAR ERROR

%{
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <linux/kd.h>
#include <linux/keyboard.h>

/* the kernel structures we want to set or print */
u_short key_map[NR_KEYMAPS][NR_KEYS];
char func_buf[FUNC_BUFSIZE];
char *func_table[NR_FUNC];
#ifdef MAX_DIACR
struct kbdiacr accent_table[MAX_DIACR];
unsigned int accent_table_size = 0;
#endif

char keymap_was_set[NR_KEYMAPS][NR_KEYS];
char *fp = func_buf;

#include "analyze.c"

#define VERSION "0.81a"

static void addkey(int index, int table, int keycode);
static void addfunc(struct kbsentry kbs_buf);
static void compose(int diacr, int base, int res);
int key_buf[16];
int mod;
%}

%%
keytable	:
		| keytable line
		;
line		: EOL
		| fullline
		| singleline
		| strline
                | compline
		;
strline		: STRING LITERAL EQUALS STRLITERAL EOL
			{
				if (KTYP($2) != KT_FN) {
					printf("'%s' is not a function key symbol\n",
						syms[KTYP($2)].table[KVAL($2)]);
					exit(1);
				}
				kbs_buf.kb_func = KVAL($2);
				addfunc(kbs_buf);
			}
		;
compline        : COMPOSE CCHAR CCHAR TO CCHAR EOL
                        {
			        compose($2, $3, $5);
			}
                ;
singleline	:	{ mod = 0; }
		  modifiers KEYCODE NUMBER EQUALS rvalue EOL
			{
				addkey($4, mod, $6);
			}
		;
modifiers	: modifiers modifier
		| modifier
		;
modifier	: SHIFT		{ mod |= (1 << KG_SHIFT);	}
		| CONTROL	{ mod |= (1 << KG_CTRL);	}
		| ALT		{ mod |= (1 << KG_ALT);		}
		| ALTGR		{ mod |= (1 << KG_ALTGR);	}
		;
fullline	: KEYCODE NUMBER EQUALS rvalue0 EOL
			{
			int i;

			for (i = 0; i < $4; i++)
				addkey($2, i, key_buf[i]);
			if ($4 == 1)
				if (KTYP(key_buf[0]) != KT_LATIN)
					for (i = 1; i < NR_KEYMAPS; i++)
						addkey($2, i, key_buf[0]);
				else if (
					(key_buf[0] >= 'a' &&
					key_buf[0] <= 'z') ||
					(key_buf[0] >= 'A' &&
					key_buf[0] <= 'Z')
				) {
	/* shift		*/	addkey($2, 1, key_buf[0] ^ 32);
	/* altgr		*/	addkey($2, 2, key_buf[0]);
	/* altgr+shift		*/	addkey($2, 3, key_buf[0] ^ 32);
	/* control		*/	addkey($2, 4, key_buf[0] & ~96);
	/* control+shift	*/	addkey($2, 5, key_buf[0] & ~96);
	/* control+altgr	*/	addkey($2, 6, key_buf[0] & ~96);
	/* control+altgr+shift	*/	addkey($2, 7, key_buf[0] & ~96);
	/* alt			*/	addkey($2, 8, K(KT_META, key_buf[0]));
	/* alt+shift		*/	addkey($2, 9, K(KT_META, key_buf[0] ^ 32));
	/* alt+altgr		*/	addkey($2, 10, K(KT_META, key_buf[0]));
	/* alt+altgr+shift	*/	addkey($2, 11, K(KT_META, key_buf[0] ^ 32));
	/* alt+control		*/	addkey($2, 12, K(KT_META, key_buf[0] & ~96));
	/* alt+control+shift	*/	addkey($2, 13, K(KT_META, key_buf[0] & ~96));
	/* alt+altgr+control	*/	addkey($2, 14, K(KT_META, key_buf[0] & ~96));
	/* alt+altgr+control+shift */	addkey($2, 15, K(KT_META, key_buf[0] & ~96));
					i = 16;
				}
			for (; i < NR_KEYMAPS; i++)
				addkey($2, i, K_HOLE);
			}
		;

rvalue0		: { $$ = 0; }
		| rvalue rvalue1
			{
			key_buf[0] = $1;
			$$ = $2 + 1;
			}
		;
rvalue1		: { $$ = 0; }
		| rvalue rvalue2
			{
			key_buf[1] = $1;
			$$ = $2 + 1;
			}
		;
rvalue2		: { $$ = 0; }
		| rvalue rvalue3
			{
			key_buf[2] = $1;
			$$ = $2 + 1;
			}
		;
rvalue3		: { $$ = 0; }
		| rvalue rvalue4
			{
			key_buf[3] = $1;
			$$ = $2 + 1;
			}
		;
rvalue4		: { $$ = 0; }
		| rvalue rvalue5
			{
			key_buf[4] = $1;
			$$ = $2 + 1;
			}
		;
rvalue5		: { $$ = 0; }
		| rvalue rvalue6
			{
			key_buf[5] = $1;
			$$ = $2 + 1;
			}
		;
rvalue6		: { $$ = 0; }
		| rvalue rvalue7
			{
			key_buf[6] = $1;
			$$ = $2 + 1;
			}
		;
rvalue7		: { $$ = 0; }
		| rvalue rvalue8
			{
			key_buf[7] = $1;
			$$ = $2 + 1;
			}
		;
rvalue8		: { $$ = 0; }
		| rvalue rvalue9
			{
			key_buf[8] = $1;
			$$ = $2 + 1;
			}
		;
rvalue9		: { $$ = 0; }
		| rvalue rvalue10
			{
			key_buf[9] = $1;
			$$ = $2 + 1;
			}
		;
rvalue10	: { $$ = 0; }
		| rvalue rvalue11
			{
			key_buf[10] = $1;
			$$ = $2 + 1;
			}
		;
rvalue11	: { $$ = 0; }
		| rvalue rvalue12
			{
			key_buf[11] = $1;
			$$ = $2 + 1;
			}
		;
rvalue12	: { $$ = 0; }
		| rvalue rvalue13
			{
			key_buf[12] = $1;
			$$ = $2 + 1;
			}
		;
rvalue13	: { $$ = 0; }
		| rvalue rvalue14
			{
			key_buf[13] = $1;
			$$ = $2 + 1;
			}
		;
rvalue14	: { $$ = 0; }
		| rvalue rvalue15
			{
			key_buf[14] = $1;
			$$ = $2 + 1;
			}
		;
rvalue15	: { $$ = 0; }
		| rvalue
			{
			key_buf[15] = $1;
			$$ = 1;
			}
		;
rvalue		: NUMBER
			{$$=$1;}
		| LITERAL
			{$$=$1;}
		;
%%			

void usage(void) {
	fprintf(stderr, "\
loadkeys version " VERSION "\

usage: loadkeys [option...] [mapfile...]

valid options are:

	-h --help	display this help text
        -m --mktable    output a \"defkeymap.c\"
        -v --verbose    report the changes
");
	exit(1);
}

char **args;
char *progname;
int optm = 0;
int verbose = 0;

int main(unsigned int argc, char *argv[]) {
	const char *short_opts = "hmv";
	const struct option long_opts[] = {
		{ "help",	no_argument, NULL, 'h' },
		{ "mktable",    no_argument, NULL, 'm' },
		{ "verbose",    no_argument, NULL, 'v' },
		{ NULL, 0, NULL, 0 }
	};
	int c;

	progname = argv[0];

	while ((c = getopt_long(argc, argv,
		short_opts, long_opts, NULL)) != -1) {
		switch (c) {
		        case 'm':
		                optm = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'h':
			case '?':
				usage();
		}
	}

	args = argv + optind - 1;
	yywrap();	/* set up the first input file, if any */
	if (yyparse()) {
		fprintf(stderr, "syntax error in map file\n");
		if(!optm)
		  fprintf(stderr, "key bindings not changed\n");
		exit(1);
	}
	if(optm)
	        mktable();
	else
	        loadkeys();
	exit(0);
}

yyerror(char *s) {
	fprintf(stderr, "%s:%d: %s\n",
		*args && strcmp(*args, "-") ? *args : "stdin", line_nr, s);
}

#undef yywrap
yywrap(void) {
	FILE *f;
	int first_file = 1; /* ugly kludge flag */

	line_nr = 1;
	if (*args)
		args++;
	if (!*args)
		return 1;
	if (!strcmp(*args, "-"))
		f = stdin;
	else if ((f = fopen(*args, "r")) == NULL) {
		fprintf(stderr, "cannot open file %s\n", *args);
		exit(1);
	}
	/*
		Can't use yyrestart if this is called before entering yyparse()
		I think assigning directly to yyin isn't necessarily safe in
		other situations, hence the flag
	*/
	if (first_file) {
		yyin = f;
		first_file = 0;
	} else
		yyrestart(f);
	return 0;
}

static void addkey(int index, int table, int keycode) {
	if (keycode == -1)
		return;
        if (index < 0 || index >= NR_KEYS) {
	        fprintf(stderr, "%s: addkeys called with bad index %d\n",
			progname, index);
		exit(1);
	}
        if (table < 0 || table >= NR_KEYMAPS) {
	        fprintf(stderr, "%s: addkeys called with bad table %d\n",
			progname, table);
		exit(1);
	}
	key_map[table][index] = keycode;
	keymap_was_set[table][index] = 1;
}

static void addfunc(struct kbsentry kbs) {
        int sh, i;
	char *p, *q, *r;

        if (kbs.kb_func >= NR_FUNC) {
	        fprintf(stderr, "%s: addfunc called with bad func %d\n",
			progname, kbs.kb_func);
		exit(1);
	}
	if ((q = func_table[kbs.kb_func])) { /* throw out old previous def */
	        sh = strlen(q) + 1;
		p = q + sh;
		while (p < fp)
		        *q++ = *p++;
		fp -= sh;
	}
	p = func_buf;                        /* find place for new def */
	for (i = 0; i < kbs.kb_func; i++)
	        if (func_table[i]) {
		        p = func_table[i];
			while(*p++);
		}
	func_table[kbs.kb_func] = p;
        sh = strlen(kbs.kb_string) + 1;
	if (fp + sh > func_buf + FUNC_BUFSIZE) {
	        fprintf(stderr, "%s: addfunc: func_buf overflow\n");
		exit(1);
	}
	q = fp;
	fp += sh;
	r = fp;
	while (q > p)
	        *--r = *--q;
	strcpy(p, kbs.kb_string);
	for (i++; i < NR_FUNC; i++)
	        if (func_table[i])
		        func_table[i] += sh;
}

#ifdef MAX_DIACR
static void compose(int diacr, int base, int res) {
        struct kbdiacr *p;
        if (accent_table_size == MAX_DIACR) {
	        fprintf(stderr, "compose table overflow\n");
		exit(1);
	}
	p = &accent_table[accent_table_size++];
	p->diacr = diacr;
	p->base = base;
	p->result = res;
}
#else
static void compose(int diacr, int base, int res){}
#endif

static int defkeys(int fd) {
	struct kbentry ke;
	int ct = 0;
	int i,j,fail;

	for(i=0; i<NR_KEYMAPS; i++) for(j=0; j<NR_KEYS; j++)
	  if (keymap_was_set[i][j]) {
	      ke.kb_index = j;
	      ke.kb_table = i;
	      ke.kb_value = key_map[i][j];
	      fail = ioctl(fd, KDSKBENT, (unsigned long)&ke);
	      if (!fail)
		ct++;
	      if(verbose)
		printf("keycode %d, table %d = %d%s\n", j, i, key_map[i][j],
		       fail ? "    FAILED" : "");
	      else if (fail)
		fprintf(stderr, "failed to bind key %d to value %d\n",
			j, key_map[i][j]);
	  }
	return ct;
}

static int deffuncs(int fd){
        int i, ct = 0;
	char *p;

        for (i = 0; i < NR_FUNC; i++)
	  if ((p = func_table[i])) {
	        kbs_buf.kb_func = i;
		strcpy(kbs_buf.kb_string, p);
		if (ioctl(fd, KDSKBSENT, (unsigned long)&kbs_buf))
		  fprintf(stderr, "failed to bind string '%s' to function %s\n",
			  kbs_buf.kb_string, syms[KT_FN].table[kbs_buf.kb_func]);
		else
		  ct++;
	    }
	return ct;
}

loadkeys () {
        int fd;
        int keyct, funcct;

	if ((fd = open("/dev/console", O_RDONLY)) < 0) {
		fprintf(stderr, "cannot open console for ioctl\n");
		exit(1);
	}

	keyct = defkeys(fd);
	funcct = deffuncs(fd);
	if (verbose) {
	        printf("\nChanged %d key%s and %d string%s\n",
		       keyct, (keyct == 1) ? "" : "s",
		       funcct, (funcct == 1) ? "" : "s");
	}
}

/*
 * mktable.c
 *
 */
outchar (unsigned char c, int comma) {
        printf("'");
        printf((c == '\'' || c == '\\') ? "\\%c" : isgraph(c) ? "%c"
	       : "\\%03o", c);
	printf(comma ? "', " : "'");
}

mktable () {
	int i, j;

	struct kbsentry kbs;
	int lentab[NR_FUNC];
	u_char *p;

	printf("/* Automatically generated by mktable */\n");
	printf("/* Do not edit this file!             */\n\n");
	printf("#include <linux/types.h>\n");
	printf("#include <linux/keyboard.h>\n");
	printf("#include <linux/kd.h>\n\n");

	printf("u_short key_map[NR_KEYMAPS][NR_KEYS] = {\n\t");
	for (i = 0; i < NR_KEYMAPS; i++) {
		printf("{");
		for (j = 0; j < NR_KEYS; j++) {
			if (!(j % 8))
				printf("\n\t");
			printf("0x%04x,\t", key_map[i][j]);
		}
		printf("\n\t}, ");
	}
	printf("\n};\n\n");

	/* make sure func_table is filled */
	for (i = 0; i < NR_FUNC; i++) if(!func_table[i]) {
	        kbs.kb_func = i;
		kbs.kb_string[0] = 0;
		addfunc(kbs);
	}

	printf("char func_buf[FUNC_BUFSIZE] = {\n");
	for (i = 0; i < NR_FUNC; i++) {
		printf("\t");
		for (p = func_table[i]; *p; p++)
		        outchar(*p, 1);
		printf("0, \n");
	}
	printf("};\n\n");

	printf("char *func_table[NR_FUNC] = {\n");
	for (i = 0; i < NR_FUNC; i++)
		printf("\tfunc_buf + %d,\n", func_table[i] - func_buf);
	printf("};\n");

#ifdef MAX_DIACR
	printf("\nstruct kbdiacr accent_table[MAX_DIACR] = {\n");
	for (i = 0; i < accent_table_size; i++) {
	        printf("\t{");
	        outchar(accent_table[i].diacr, 1);
		outchar(accent_table[i].base, 1);
		outchar(accent_table[i].result, 0);
		printf("},");
		if(i%2) printf("\n");
	}
	if(i%2) printf("\n");
	printf("};\n\n");
	printf("unsigned int accent_table_size = %d;\n",
	       accent_table_size);
#endif /* MAXDIACR */

	exit(0);
}


